/*
 * Copyright 2013 Mark Slater
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package net.sourceforge.writexml;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;
import java.io.Writer;

import static net.sourceforge.writexml.Document.document;

/**
 * Formatter for outputting {@code Document}s and {@code Tag}s as XML to a {@code Writer}.
 */
public final class CompactXmlFormatter {

    public CompactXmlFormatter() {
    }

    /**
     * Writes the given {@code Tag} as the root of an XML document to the given {@code Writer}.
     *
     * @param rootTag the {@code Tag} that represents the root of the document.
     * @param writer  the {@code Writer} to output to.
     * @throws XmlWriteException in the event of a failure outputting to the {@code Writer}.
     */
    public void write(final Tag rootTag, final Writer writer) throws XmlWriteException {
        write(document(rootTag), writer);
    }

    /**
     * Writes the given {@code Document} as an XML document to the given {@code Writer}.
     *
     * @param document the {@code Document} to write.
     * @param writer   the {@code Writer} to output to.
     * @throws XmlWriteException in the event of a failure outputting to the {@code Writer}.
     */
    public void write(final Document document, final Writer writer) throws XmlWriteException {
        StreamResult streamResult = new StreamResult(writer);
        final TransformerHandler transformerHandler;
        try {
            transformerHandler = ((SAXTransformerFactory) SAXTransformerFactory.newInstance()).newTransformerHandler();
        } catch (TransformerConfigurationException e) {
            throw new XmlWriteException("Failed configuring CompactXmlFormatter", e);
        }
        transformerHandler.setResult(streamResult);
        document.writeTo(new XmlWriter() {
            public void write(final Tag child) throws XmlWriteException {
                child.writeTo(this);
            }

            private org.xml.sax.Attributes asXmlAttributes(final Attributes attributes) {
                final AttributesImpl resultAttributes = new AttributesImpl();
                for (Attribute attribute : attributes) {
                    resultAttributes.addAttribute(
                            attribute.namespace().asString(),
                            attribute.name().value,
                            (attribute.namespace() == Namespace.XML ? "xml:" : "") + attribute.name().value,
                            null,
                            attribute.value().value);
                }
                return resultAttributes;
            }

            @Override
            public void writeStartDocument() throws XmlWriteException {
                try {
                    transformerHandler.startDocument();
                } catch (SAXException e) {
                    throw new XmlWriteException("Failed writing start document to XML", e);
                }
            }

            @Override
            public void writeText(final String text) throws XmlWriteException {
                try {
                    transformerHandler.characters(text.toCharArray(), 0, text.length());
                } catch (SAXException e) {
                    throw new XmlWriteException("Failed writing text [" + text + "] to XML", e);
                }
            }

            @Override
            public void writeCdata(final String cdata) throws XmlWriteException {
                try {
                    transformerHandler.startCDATA();
                    transformerHandler.characters(cdata.toCharArray(), 0, cdata.length());
                } catch (final SAXException e) {
                    throw new XmlWriteException("Failed writing CDATA section [" + cdata + "] to XML", e);
                }
            }

            @Override
            public void writeStartTag(final Namespace namespace, final TagName name, final Attributes attributes) throws XmlWriteException {
                try {
                    transformerHandler.startElement(namespace.asString(), null, name.asString(), asXmlAttributes(attributes));
                } catch (SAXException e) {
                    throw new XmlWriteException("Failed writing an open tag named [" + name + "] to XML", e);
                }
            }

            @Override
            public void writeEndTag(final Namespace namespace, final TagName name) throws XmlWriteException {
                try {
                    transformerHandler.endElement(namespace.asString(), null, name.asString());
                } catch (SAXException e) {
                    throw new XmlWriteException("Failed writing a close tag named [" + name + "] to XML", e);
                }
            }

            @Override
            public void writeEndDocument() throws XmlWriteException {
                try {
                    transformerHandler.endDocument();
                } catch (SAXException e) {
                    throw new XmlWriteException("Failed writing end document to XML", e);
                }
            }

            @Override
            public void writeDtd(final DtdName name, final DtdPublicId publicId, final DtdSystemId systemId) throws XmlWriteException {
                try {
                    transformerHandler.startDTD(name.value, publicId.value, systemId.value);
                    transformerHandler.endDTD();
                } catch (SAXException e) {
                    throw new XmlWriteException("Failed writing DTD to XML", e);
                }
            }
        });
    }

}
