/*
 * Copyright 2013 Mark Slater
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package net.sourceforge.writexml;

import static net.sourceforge.writexml.ObjectUtils.checkNotNull;

/**
 * An XML document - the root object for output as XML.
 * <p/>
 * Threadsafe and immutable.
 */
public abstract class Document {

    private Document() {
    }

    /**
     * Factory method for a document without a DTD.
     *
     * @param rootTag the root of the document.
     * @return a {@code Document} with the given root.
     */
    public static Document document(final Tag rootTag) {
        checkNotNull(rootTag, "Cannot instantiate Document with null root tag");
        return new Document() {
            @Override
            void writeTo(final XmlWriter xmlWriter) throws XmlWriteException {
                xmlWriter.writeStartDocument();
                xmlWriter.write(rootTag);
                xmlWriter.writeEndDocument();
            }
        };
    }

    /**
     * Factory method for a document with a DTD.
     *
     * @param dtd     the DTD of the document.
     * @param rootTag the root of the document.
     * @return a {@code Document} with the given DTD and root.
     */
    public static Document document(final Dtd dtd, final Tag rootTag) {
        checkNotNull(dtd, "Cannot instantiate Document with null DTD");
        checkNotNull(rootTag, "Cannot instantiate Document with null root tag");
        return new Document() {
            @Override
            void writeTo(final XmlWriter xmlWriter) throws XmlWriteException {
                dtd.writeTo(xmlWriter);
                xmlWriter.writeStartDocument();
                xmlWriter.write(rootTag);
                xmlWriter.writeEndDocument();
            }
        };
    }

    abstract void writeTo(XmlWriter xmlWriter) throws XmlWriteException;
}
